from .learner import Learner, LearnerMAS
from .policy_gradient import PG_Learner
from .policy_gradient import A2C_Learner
from .policy_gradient import PPOCLIP_Learner
from .policy_gradient import PPOKL_Learner
from .policy_gradient import PPG_Learner
from .policy_gradient import DDPG_Learner
from .policy_gradient import TD3_Learner
from .policy_gradient import SAC_Learner
from .policy_gradient import SACDIS_Learner
from .policy_gradient import PDQN_Learner
from .policy_gradient import MPDQN_Learner
from .policy_gradient import SPDQN_Learner

from .qlearning_family import DQN_Learner
from .qlearning_family import DuelDQN_Learner
from .qlearning_family import DDQN_Learner
from .qlearning_family import PerDQN_Learner
from .qlearning_family import C51_Learner
from .qlearning_family import QRDQN_Learner
from .qlearning_family import DRQN_Learner

from .multi_agent_rl import IQL_Learner
from .multi_agent_rl import VDN_Learner
from .multi_agent_rl import QMIX_Learner
from .multi_agent_rl import WQMIX_Learner
from .multi_agent_rl import QTRAN_Learner
from .multi_agent_rl import VDAC_Learner
from .multi_agent_rl import COMA_Learner
from .multi_agent_rl import IDDPG_Learner
from .multi_agent_rl import MADDPG_Learner
from .multi_agent_rl import MFQ_Learner
from .multi_agent_rl import MFAC_Learner
from .multi_agent_rl import IPPO_Learner
from .multi_agent_rl import MAPPO_Clip_Learner
from .multi_agent_rl import ISAC_Learner
from .multi_agent_rl import MASAC_Learner
from .multi_agent_rl import MATD3_Learner
