method_list = {
    'a2c': ['box2d', 'classic_control', 'atari', 'mujoco'],
    'c51': ['box2d', 'classic_control', 'atari'],
    'coma': ['mpe', 'sc2'],
    'dcg': ['mpe', 'sc2'],
    'ddpg': ['box2d', 'classic_control', 'mujoco'],
    'ddqn': ['box2d', 'classic_control', 'atari'],
    'dqn': ['box2d', 'classic_control', 'atari'],
    'drqn': ['box2d', 'classic_control', 'atari'],
    'dueldqn': ['box2d', 'classic_control', 'atari'],
    'iddpg': ['mpe'],
    'ippo': ['football', 'mpe', 'sc2'],
    'iql': ['football', 'magent2', 'mpe', 'sc2'],
    'isac': ['mpe'],
    'maddpg': ['mpe'],
    'mappo': ['football', 'mpe', 'sc2'],
    'masac': ['mpe'],
    'matd3': ['mpe'],
    'mfac': ['mpe'],
    'mfq': ['magent2', 'mpe'],
    'mpdqn': ['Platform'],
    'noisydqn': ['box2d', 'classic_control', 'atari'],
    'pdqn': ['Platform'],
    'perdqn': ['box2d', 'classic_control', 'atari'],
    'pg': ['box2d', 'classic_control', 'mujoco'],
    'ppg': ['box2d', 'classic_control', 'mujoco'],
    'ppo': ['box2d', 'classic_control', 'atari', 'mujoco'],
    'qmix': ['football', 'mpe', 'sc2'],
    'qrdqn': ['box2d', 'classic_control', 'atari'],
    'qtran': ['mpe'],
    'random': ['mpe'],
    'sac': ['box2d', 'classic_control', 'atari', 'mujoco'],
    'spdqn': ['Platform'],
    'td3': ['box2d', 'classic_control', 'mujoco'],
    'vdac': ['mpe', 'sc2'],
    'vdn': ['football', 'mpe', 'sc2'],
    'wqmix': ['mpe', 'sc2']
}
